/* package which supports access to the berkeley db package.

   dbopen options name file_name - creates a database object.  options
   	can include -hash, -btree, -create, -readonly

   possible operations on a database object include:

	db close

	db names
	db -glob names pattern

	db delete key 
	db -glob delete pattern

	db delete_cursor

	db get key
	db -glob get pattern

	db put ?options? key value
		options can include -cursor or -setcursor -nooverwrite

	db sync

	db foreach key value block
	db -glob foreach pattern key value block

*/

#include <db.h>
#include <tcl.h>
#include <sys/types.h>
#include <limits.h>
#include <fcntl.h>
#include <string.h>
#include <malloc.h>

#include "berk.h"
#include "binary.h"

typedef struct {
    DB *db;
    int binary;
} db_clientdata;
    

/* returns true if a character is special for glob style wildcards */
static int glob_special(int ch)
{
    switch (ch) {
      case '*':
      case '?':
      case '[':
	  return 1;
      default:
	  return 0;
    }
}

/* returns the prefix of a string which doesn't include any
   wildcard magic in it.  it is guaranteed that all strings which
   match the wildcard will start with this prefix.

   the prefix is returned in newly malloc'ed space, the length
   of the prefix is returned through length if length is not
   NULL */
static char *wildcard_prefix(char *pattern, int *length)
{
    int i, j;
    char *prefix;

    /* duplicate the string.  this is guaranteed to allocate enough space */
    prefix = strdup(pattern);
    j = 0;

    /* copy the invariant prefix, stop at the first magic character */
    for (i=0;prefix[i] && !glob_special(prefix[i]);i++) {
	/* undo escaped characters as we go */
	if (prefix[i] == '\\' && prefix[i+1]) {
	    prefix[j] = prefix[i+1];
	    i++;
	    j++;
	}
	else {
	    prefix[j] = prefix[i];
	    j++;
	}
    }
    if (length) *length = j;
    prefix[j] = 0;

    return prefix;
}

/* this procedure handles all of the calls to any database object.

   the syntax supported includes the following forms:

    db close
	closes the database object and reclaims space.

    db names
	returns a list of all the keys in the database

    db -glob names pattern
    db names pattern
	returns a list of all the keys in the database which match the pattern
    
    db delete key 
	deletes the first record with the specified key
    
    db -glob delete pattern
	deletes all records whose keys match the pattern
    
    db get key
	gets the value associated with the specified key

    db -glob get pattern
	gets a list containing all the values of records whose keys match the
	pattern
    
    db put key value
	insert a new record or overwrite an old record.
    
    db sync
	flush all pending writes to disk
    
    db foreach key value block
	loop through all records in the database.  for each record,
	key will be assigned the key and value will be assigned the
	value and then block will be executed.  this works just like
	the tcl foreach loop:

	    foreach key [db names] value [db -glob get *] block

	the only difference is that the form provided here is much
	faster, requires *much* less memory and can exit before all
	values in the database have been processed.

    db -glob foreach pattern key value block
	same as the previous command, but only records whose keys
	match the pattern are used.

 */
static int db_methods(ClientData cl, Tcl_Interp *interp,
		      int argc, char *argv[])
{
    int glob = 0;
    int flag = 0;

    db_clientdata *db;
    DBT key, value;

    Tcl_DString r;


    /* get the database from the client data */
    db = (db_clientdata *) cl;

    if (!db || !db->db) {
	Tcl_SetResult(interp,
		      "Can't happen, but I got a null internal pointer",
		      TCL_STATIC);
	return TCL_ERROR;
    }

    /* first check for the glob option which can apply to lots of commands */
    while (argc > 1 && argv[1][0] == '-') {
	if (strcmp(argv[1], "-glob") == 0) {
	    glob = 1;
	
	}
	else if (strcmp(argv[1], "-nodup") == 0) {
	    flag = R_NOOVERWRITE;
	}
	else {
	    Tcl_AppendResult(interp, "bad option, wanted -glob, -nodup or nothing before db command",
			     NULL);
	    return TCL_ERROR;
	}
	/* pretend we have fewer arguments than we really had */
	argc--;
	argv++;
	argv[0] = argv[-1];
    }

    /* the shortest legal expression is "db names" */
    if (argc < 2) {
	Tcl_SetResult(interp, "bad option for db.  must be one of names, close, delete, get, put, sync or foreach", TCL_STATIC);
	return TCL_ERROR;
    }

    /* now check for each sub-command.  using strcmp on each possibility is no
       speed penalty since the database access is sooo much slower anyway */
    if (strcmp(argv[1], "close") == 0) {
	if (glob) {
	    Tcl_AppendResult(interp, "-glob doesn't make sense with close",
			     NULL);
	    return TCL_ERROR;
	}
	if (flag) {
	    Tcl_AppendResult(interp, "-nodup doesn't make sense with close",
			     NULL);
	    return TCL_ERROR;
	}

	/* the database will be closed as part of the delete operation */
	Tcl_DeleteCommand(interp, argv[0]);
    }
    else if (strcmp(argv[1], "delete") == 0) {
	if (argc != 3) {
	    Tcl_SetResult(interp, "need key for db delete command",
			  TCL_STATIC);
	    return TCL_ERROR;
	}
	if (flag) {
	    Tcl_AppendResult(interp, "-nodup doesn't make sense with delete",
			     NULL);
	    return TCL_ERROR;
	}
	if (glob) {
	    int flag;
	    int len, rv;
	    char *prefix;

	    if (argc < 3) {
		Tcl_AppendResult(interp, "need pattern for wildcard delete ",
				 argv[0], NULL);
		return TCL_ERROR;
	    }

	    /* first we get the prefix so we know where to start */
	    prefix = wildcard_prefix(argv[2], &len);
	    key.data = prefix;
	    key.size = len;
	    /* the berkeley hash package can't seem to handle an empty string
	       real well.  with this code, we give it a break */
	    if (len) flag = R_CURSOR;
	    else flag = R_FIRST;

	    Tcl_ResetResult(interp);
	    /* this for loop starts just after the specified key and could
	       concievably look at all of the database */
	    for (rv=db->db->seq(db->db, &key, &value, flag);
		 rv == 0;
		 rv=db->db->seq(db->db, &key, &value, R_NEXT)) {

		/* if the key we found matches our pattern, we do the deed */
		if (Tcl_StringMatch(key.data, argv[2])) {
		    db->db->del(db->db, &key, R_CURSOR);
		}
		/* if the key that we found has a different prefix, then we bail */
		if (strncmp(prefix, key.data, len) < 0) break;
	    }

	    /* if we blew out with a database error, then we must note the problem */
	    if (rv == -1) {
		Tcl_ResetResult(interp);
		Tcl_AppendResult(interp, "error getting data from ",
				 argv[0], NULL);
		return TCL_ERROR;
	    }

	    /* remember?  we allocated this in wildcard_prefix */
	    free(prefix);
	    return TCL_OK;
	}
	else {
	    /* delete for a known key is much simpler than the wildcard operation */
	    key.data = argv[2];
	    key.size = strlen(argv[2])+1;
	    switch (db->db->del(db->db, &key, 0)) {
	      case 0:
		  break;
	      case 1:
		  /* no error, but the record wasn't found */
		  Tcl_ResetResult(interp);
		  Tcl_AppendResult(interp, "key \"", argv[2], "\" not found in ",
				   argv[0], NULL);
		  return TCL_ERROR;
		  break;
	      case -1:
		  Tcl_ResetResult(interp);
		  Tcl_AppendResult(interp, "error encountered deleting \"",
				   argv[2], "\" from ", argv[0], NULL);
		  return TCL_ERROR;
		  break;
	    }

	    return TCL_OK;
	}
    }
    else if (strcmp(argv[1], "get") == 0) {
	if (argc != 3) {
	    Tcl_SetResult(interp, "need key for db get command",
			  TCL_STATIC);
	    return TCL_ERROR;
	}

	if (flag) {
	    Tcl_AppendResult(interp, "-nodup doesn't make sense with get",
			     NULL);
	    return TCL_ERROR;
	}
	if (glob) {
	    int len, rv;
	    char *prefix;

	    if (argc < 3) {
		Tcl_AppendResult(interp, "need pattern for wildcard get ",
				 argv[0], NULL);
		return TCL_ERROR;
	    }

	    /* here we get the invariant prefix so that we can examine fewer
	       records */
	    prefix = wildcard_prefix(argv[2], &len);
	    key.data = prefix;
	    key.size = len;

	    /* the berkeley hash package can't seem to handle an empty string
	       real well.  with this code, we give it a break */
	    if (len) flag = R_CURSOR;
	    else flag = R_FIRST;

	    Tcl_ResetResult(interp);
	    Tcl_DStringInit(&r);

	    /* this for loop starts at the first key at or after the
	       given prefix and could continue through the entire database */
	    for (rv=db->db->seq(db->db, &key, &value, flag);
		 rv == 0;
		 rv=db->db->seq(db->db, &key, &value, R_NEXT)) {
		/* all matching items are added to our result */
		if (Tcl_StringMatch(key.data, argv[2])) {
		    if (db->binary) {
			Bin_DStringAppendElement(&r, value.data, value.size);
		    }
		    else {
			Tcl_DStringAppendElement(&r, value.data);
		    }
		}
		/* if we find a change in the prefix, we blow out */
		if (strncmp(prefix, key.data, len) < 0) break;
	    }

	    /* if we exited the loop due to error, we need to report that */
	    if (rv == -1) {
		Tcl_DStringFree(&r);
		Tcl_ResetResult(interp);
		Tcl_AppendResult(interp, "error getting data from ",
				 argv[0], NULL);
		return TCL_ERROR;
	    }

	    /* we allocated this in wildcard_prefix */
	    free(prefix);
	    Tcl_DStringResult(interp, &r);
	    return TCL_OK;
	}
	else {
	    /* non-wildcard get is much simpler... we just do the operation
	       and report the results */
	    key.data = argv[2];
	    key.size = strlen(argv[2])+1;
	    switch (db->db->get(db->db, &key, &value, 0)) {
	      case 0:
		  break;
	      case 1:
		  Tcl_ResetResult(interp);
		  Tcl_AppendResult(interp, "key \"", argv[2],
				   "\" couldn't be found in ", argv[0], NULL);
		  return TCL_ERROR;
		  break;
	      case -1:
		  Tcl_ResetResult(interp);
		  Tcl_AppendResult(interp, "error encountered finding \"",
				   argv[2], "\" in ", argv[0], NULL);
		  return TCL_ERROR;
		  break;
	    }
	    if (db->binary) {
		Bin_SetResult(interp, value.data, value.size);
	    }
	    else {
		Tcl_SetResult(interp, value.data, TCL_VOLATILE);
	    }
	    return TCL_OK;
	}
    }
    else if (strcmp(argv[1], "put") == 0) {
	if (glob) {
	    Tcl_AppendResult(interp, "-glob doesn't make sense with put",
			     NULL);
	    return TCL_ERROR;
	}

	if (argc != 4) {
	    Tcl_SetResult(interp, "need key for db put command",
			  TCL_STATIC);
	    return TCL_ERROR;
	}

	/* we can't have a wildcard put, so this is pretty simple code */
	key.data = argv[2];
	key.size = strlen(argv[2])+1;
	value.data = argv[3];
	if (db->binary) {
	    value.size = Bin_DecodeData(value.data, value.data, -1);
	}
	else {
	    value.size = strlen(argv[3])+1;
	}

	switch (db->db->put(db->db, &key, &value, flag)) {
	  case 0:
	      break;
	  case 1:
	      Tcl_ResetResult(interp);
	      Tcl_AppendResult(interp, "key \"", argv[2],
			       "\" couldn't be added to ", argv[0], NULL);
	      return TCL_ERROR;
	      break;
	  case -1:
	      Tcl_ResetResult(interp);
	      Tcl_AppendResult(interp, "error encountered adding \"",
			       argv[2], "\" in ", argv[0], NULL);
	      return TCL_ERROR;
	      break;
	}
	Tcl_ResetResult(interp);
	return TCL_OK;
    }
    else if (strcmp(argv[1], "sync") == 0) {
	if (glob) {
	    Tcl_AppendResult(interp, "-glob doesn't make sense with sync",
			     NULL);
	    return TCL_ERROR;
	}

	/* sync is a no-brainer */
	db->db->sync(db->db, 0);
    }
    else if (strcmp(argv[1], "foreach") == 0) {
	int len, rv;
	int start_flag;
	char *keyvar, *valvar;
	char *prefix, *pattern, *command;

	if (flag) {
	    Tcl_AppendResult(interp, "-nodup doesn't make sense with foreach",
			     NULL);
	    return TCL_ERROR;
	}

	/* this foreach interpreter is parametrized since we do a wildcard
	   scan even if the user didn't specify a pattern */
	if (glob) {
	    /* with globbing we need to get their pattern */
	    if (argc != 6) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
				 argv[0], " -glob foreach pattern ",
				 "key_variable value_variable command\"",
				 NULL);
		return TCL_ERROR;
	    }

	    pattern = argv[2];
	    keyvar = argv[3];
	    valvar = argv[4];
	    command = argv[5];

	}
	else {
	    if (argc != 5) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
				 argv[0], " foreach ",
				 "key_variable value_variable command\"",
				 NULL);
		return TCL_ERROR;
	    }

	    /* without globbing, the pattern matches everything */
	    pattern = "*";
	
	    keyvar = argv[2];
	    valvar = argv[3];
	    command = argv[4];
	}

	prefix = wildcard_prefix(pattern, &len);
	key.data = prefix;
	key.size = len;
	/* with any sort of globbing, we start at the prefix, except that
	   the berkeley hash package can't seem to handle an empty string
	   real well.  with this code, we give it a break */
	if (len) start_flag = R_CURSOR;
	else start_flag = R_FIRST;

	Tcl_ResetResult(interp);
	for (rv=db->db->seq(db->db, &key, &value, start_flag);
	     rv == 0;
	     rv=db->db->seq(db->db, &key, &value, R_NEXT)) {
	    if (Tcl_StringMatch(key.data, pattern)) {
		if (Tcl_SetVar(interp, keyvar, key.data,
			       TCL_LEAVE_ERR_MSG) == NULL) {
		    return TCL_ERROR;
		}
		if (Tcl_SetVar(interp, valvar, value.data,
			       TCL_LEAVE_ERR_MSG) == NULL) {
		    return TCL_ERROR;
		}

		/* after we set the loop variables, then we need to evaluate the
		   loop body.  if it does a break or error then we need to blow
		   out in the right way */
		switch (Tcl_Eval(interp, command)) {
		  case TCL_ERROR:
		      Tcl_ResetResult(interp);
		      Tcl_AppendResult(interp, "error iterating through ",
				       argv[0], NULL);
		      return TCL_ERROR;
		  case TCL_BREAK:
		  case TCL_RETURN:
		      return TCL_OK;
		  case TCL_CONTINUE:
		  case TCL_OK:
		      break;
		}
	    }
	    /* if the prefix changes, then we need to quit looping */
	    if (strncmp(prefix, key.data, len) < 0) break;
	}
	if (rv == -1) {
	    Tcl_ResetResult(interp);
	    Tcl_AppendResult(interp, "error iterating through ",
			     argv[0], NULL);
	    return TCL_ERROR;
	}
	free(prefix);
	return TCL_OK;
    }
    else if (strcmp(argv[1], "names") == 0) {
	int flag;
	int rv, len;

	/* this is a bit confusing.  if glob is set, then the user is
	   using syntax of the form "db -glob names pattern" and at
	   this point, argc must be 3.  on the other hand, the new
	   syntax "db names pattern" would also get us here with
	   argc==3.  since we want to support both old and new syntax,
	   and still have full error checking, this if condition looks
	   a bit redundant. */
	if (glob || argc == 3) {
	    char *prefix;

	    if (argc != 3) {
		Tcl_AppendResult(interp, "need pattern for wildcard scan ",
				 argv[0], NULL);
		return TCL_ERROR;
	    }

	    prefix = wildcard_prefix(argv[2], &len);
	    key.data = prefix;
	    key.size = len;
	
	    /* the berkeley hash package can't seem to handle an empty string
	       real well.  with this code, we give it a break */
	    if (len) flag = R_CURSOR;
	    else flag = R_FIRST;

	    Tcl_ResetResult(interp);
	    for (rv=db->db->seq(db->db, &key, &value, flag);
		 rv == 0;
		 rv=db->db->seq(db->db, &key, &value, R_NEXT)) {
		if (Tcl_StringMatch(key.data, argv[2])) {
		    Tcl_AppendElement(interp, key.data);
		}
		if (strncmp(prefix, key.data, len) < 0) break;
	    }
	    if (rv == -1) {
		Tcl_ResetResult(interp);
		Tcl_AppendResult(interp, "error getting names from ",
				 argv[0], NULL);
		return TCL_ERROR;
	    }
	    free(prefix);
	    return TCL_OK;
	}
	else {
	    if (argc != 2) {
		Tcl_AppendResult(interp, "bad arguments, wanted \"",
				 argv[0], " ?-glob? names\"", NULL);
		return TCL_ERROR;
	    }
	    Tcl_ResetResult(interp);
	    for (rv=db->db->seq(db->db, &key, &value, R_FIRST);
		 rv == 0;
		 rv=db->db->seq(db->db, &key, &value, R_NEXT)) {
		Tcl_AppendElement(interp, key.data);
	    }
	    if (rv == -1) {
		Tcl_ResetResult(interp);
		Tcl_AppendResult(interp, "error getting names from ",
				 argv[0], NULL);
		return TCL_ERROR;
	    }
	    return TCL_OK;
	}
    }
    else {
	Tcl_SetResult(interp, "bad usage, wanted one of:

    db close
	closes the database object and reclaims space.

    db names
	returns a list of all the keys in the database

    db -glob names pattern
	returns a list of all the keys in the database which match the pattern
    
    db delete key 
	deletes the first record with the specified key
    
    db -glob delete pattern
	deletes all records whose keys match the pattern
    
    db get key
	gets the value associated with the specified key

    db -glob get pattern
	gets a list containing all the values of records whose keys match the
	pattern
    
    db put key value
	insert a new record or overwrite an old record.
    
    db sync
	flush all pending writes to disk
    
    db foreach key value block
	loop through all records in the database.  for each record,
	key will be assigned the key and value will be assigned the
	value and then block will be executed.  this works just like
	the tcl foreach loop:

	    foreach key [db names] value [db -glob get *] block

	the only difference is that the form provided here is much
	faster, requires *much* less memory and can exit before all
	values in the database have been processed.

    db -glob foreach pattern key value block
	same as the previous command, but only records whose keys
	match the pattern are used.", TCL_STATIC);
	return TCL_ERROR;
    }
    return TCL_OK;
}

/* deallocating a db object merely requires that we close the underlying
   database */
static void db_delete(ClientData cl)
{
    db_clientdata *db;

    db = (db_clientdata *) cl;
    if (db && db->db) db->db->close(db->db);
}


/* creating a database can be done with any of these syntaxes:

   dbopen ?-create? ?-new? ?-binary? db_name

   the first form opens an in-memory database which is not associated
   with any file and which will not persist beyond the life of this
   process.

   the second form opens a pre-existing database which is stored in
   file_name.  it fails if the file does not already contain a valid
   database.  the database is opened read/write if possible and
   read-only if necessary.

   the third form opens a new database to be stored in file_name.  if
   a file with that name already exists, then it opens that database.
   if the pre-existing file cannot be opened read/write, then the open
   fails.

   the fourth form opens a new database and fails if a file of that name
   already exists.  this is done in such a way as to avoid all race
   conditions.

   */

static int db_open(ClientData cl, Tcl_Interp *interp,
		   int argc, char *argv[])
{
    int new;
    char *fname;
    db_clientdata *db;

    BTREEINFO db_options;

    memset(&db_options, 0, sizeof(db_options));

    new = 0;

    /* without a file name, we open an in-memory database with
       no attached file */
    db = malloc(sizeof(*db));
    db->db = NULL;
    db->binary = 0;
    if (!db) {
	Tcl_SetResult(interp, "Cannot allocate memory", TCL_STATIC);
	return TCL_ERROR;
    }

    if (argc == 2) {
	fname = NULL;
    }
    else {
	new = O_RDWR;
	db->binary = 0;
	while (argc > 1 && argv[1][0] == '-') {
	    if (argc > 1 && strcmp(argv[1], "-create") == 0) {
		new |= O_CREAT;
	    }
	    else if (argc > 1 && strcmp(argv[1], "-new") == 0) {
		new |= O_CREAT | O_EXCL;
	    }
	    else if (argc > 1 && strcmp(argv[1], "-readonly") == 0) {
		new &= O_RDWR;
		new |= O_RDONLY;
	    }
	    else if (argc > 1 && strcmp(argv[1], "-binary") == 0) {
		db->binary = 1;
	    }
	    else if (argc > 1 && strcmp(argv[1], "-multiple") == 0) {
		db_options.flags |= R_DUP;
	    }
	    else if (argc > 1 && strcmp(argv[1], "-cache") == 0) {
		if (argc < 3) {
		    Tcl_SetResult(interp, "need cache size with -cache option",
				  TCL_STATIC);
		    return TCL_ERROR;
		}
		if (Tcl_GetInt(interp, argv[2], &db_options.cachesize) != TCL_OK) {
		    return TCL_ERROR;
		}
		argc--;
		argv++;
		argv[0] = argv[-1];
	    }
	    
	    argc--;
	    argv++;
	    argv[0] = argv[-1];
	}
    
	if (argc != 2 && argc != 3) {
	    Tcl_AppendResult(interp,
			     "bad arguments, needed \"dbopen ?-readonly? ?-new? ?-create? ",
			     "db_name ?file_name?\"", NULL);
	    return TCL_ERROR;
	}
	fname = argv[2];

	/* open the database */
	db->db = dbopen(fname, new | O_RDWR, 0666, DB_BTREE, &db_options);

	/* didn't work? */
	if (db->db == NULL) {
	    /* for a pre-existing file, try it read-only if we need to */
	    if (!new) {
		db->db = dbopen(argv[2], O_RDONLY, 0666, DB_BTREE, NULL);
	    }
	    if (db->db == NULL) {
		Tcl_AppendResult(interp, "cannot open database file ",
				 argv[2], NULL);
	    }
	    return TCL_ERROR;
	}
	Tcl_CreateCommand(interp, argv[1], db_methods, db, db_delete);
    }
    return TCL_OK;
}


int Berk_Init(Tcl_Interp *interp)
{
    Tcl_CreateCommand(interp, "dbopen", db_open, NULL, NULL);
    return TCL_OK;
}
